@@ -13,10 +13,10 @@
 --
 -----------------------------------------------------------------------------
 
-module XMonad.Main (xmonad) where
+module XMonad.Main (xmonad, launch) where
 
 import System.Locale.SetLocale
-import Control.Arrow (second)
+import qualified Control.Exception.Extensible as E
 import Data.Bits
 import Data.List ((\\))
 import Data.Function
@@ -37,7 +37,7 @@ import qualified XMonad.StackSet as W
 import XMonad.Operations
 
 import System.IO
-
+import System.Directory
 import System.Info
 import System.Environment
 import System.Posix.Process (executeFile)
@@ -59,33 +59,29 @@ xmonad :: (LayoutClass l Window, Read (l Window)) => XConfig l -> IO ()
 xmonad conf = do
     installSignalHandlers -- important to ignore SIGCHLD to avoid zombies
 
-    let launch serializedWinset serializedExtState args = do
+    let launch' args = do
               catchIO buildLaunch
               conf' @ XConfig { layoutHook = Layout l }
                   <- handleExtraArgs conf args conf{ layoutHook = Layout (layoutHook conf) }
-              withArgs [] $
-                xmonadNoargs (conf' { layoutHook = l })
-                              serializedWinset
-                              serializedExtState
+              withArgs [] $ launch (conf' { layoutHook = l })
 
     args <- getArgs
     case args of
-        ("--resume": ws : xs : args') -> launch (Just ws) (Just xs) args'
+        ("--resume": ws : xs : args') -> migrateState ws xs >> launch' args'
         ["--help"]            -> usage
         ["--recompile"]       -> recompile True >>= flip unless exitFailure
         ["--restart"]         -> sendRestart
         ["--version"]         -> putStrLn $ unwords shortVersion
         ["--verbose-version"] -> putStrLn . unwords $ shortVersion ++ longVersion
-        "--replace" : args'   -> do
-                                  sendReplace
-                                  launch Nothing Nothing args'
-        _                     -> launch Nothing Nothing args
+        "--replace" : args'   -> sendReplace >> launch' args'
+        _                     -> launch' args
  where
     shortVersion = ["xmonad", showVersion version]
     longVersion  = [ "compiled by", compilerName, showVersion compilerVersion
                    , "for",  arch ++ "-" ++ os
                    , "\nXinerama:", show compiledWithXinerama ]
 
+
 usage :: IO ()
 usage = do
     self <- getProgName
@@ -99,13 +95,13 @@ usage = do
         "  --restart                    Request a running xmonad process to restart" :
         []
 
--- | Build "~\/.xmonad\/xmonad.hs" with ghc, then execute it.  If there are no
--- errors, this function does not return.  An exception is raised in any of
--- these cases:
+-- | Build the xmonad configuration file with ghc, then execute it.
+-- If there are no errors, this function does not return.  An
+-- exception is raised in any of these cases:
 --
 --   * ghc missing
 --
---   * both "~\/.xmonad\/xmonad.hs" and "~\/.xmonad\/xmonad-$arch-$os" missing
+--   * both the configuration file and executable are missing
 --
 --   * xmonad.hs fails to compile
 --
@@ -118,7 +114,7 @@ usage = do
 buildLaunch ::  IO ()
 buildLaunch = do
     recompile False
-    dir  <- getXMonadDir
+    dir  <- getXMonadDataDir
     args <- getArgs
     whoami <- getProgName
     let compiledConfig = "xmonad-"++arch++"-"++os
@@ -144,15 +140,28 @@ sendReplace = do
     rootw  <- rootWindow dpy dflt
     replace dpy dflt rootw
 
-
--- |
--- The main entry point
+-- | Entry point into xmonad for custom builds.
 --
-xmonadNoargs :: (LayoutClass l Window, Read (l Window)) => XConfig l
-    -> Maybe String -- ^ serialized windowset
-    -> Maybe String -- ^ serialized extensible state
-    -> IO ()
-xmonadNoargs initxmc serializedWinset serializedExtstate = do
+-- This function isn't meant to be called by the typical xmonad user
+-- because it:
+--
+--   * Does not process any command line arguments.
+--
+--   * Therefore doesn't know how to restart a running xmonad.
+--
+--   * Does not compile your configuration file since it assumes it's
+--     actually running from within your compiled configuration.
+--
+-- Unless you know what you are doing, you should probably be using
+-- the 'xmonad' function instead.
+--
+-- However, if you are using a custom build environment (such as
+-- stack, cabal, make, etc.) you will likely want to call this
+-- function instead of 'xmonad'.  You probably also want to have a key
+-- binding to the 'XMonad.Operations.restart` function that restarts
+-- your custom binary with the resume flag set to @True@.
+launch :: (LayoutClass l Window, Read (l Window)) => XConfig l -> IO ()
+launch initxmc = do
     -- setup locale information from environment
     setLocale LC_ALL (Just "")
     -- ignore SIGPIPE and SIGCHLD
@@ -176,6 +185,7 @@ xmonadNoargs initxmc serializedWinset serializedExtstate = do
     xSetErrorHandler -- in C, I'm too lazy to write the binding: dons
 
     xinesc <- getCleanedScreenInfo dpy
+
     nbc    <- do v            <- initColor dpy $ normalBorderColor  xmc
                  ~(Just nbc_) <- initColor dpy $ normalBorderColor Default.def
                  return (fromMaybe nbc_ v)
@@ -190,19 +200,6 @@ xmonadNoargs initxmc serializedWinset serializedExtstate = do
         lreads = readsLayout layout
         initialWinset = let padToLen n xs = take (max n (length xs)) $ xs ++ repeat ""
             in new layout (padToLen (length xinesc) (workspaces xmc)) $ map SD xinesc
-        maybeRead reads' s = case reads' s of
-                                [(x, "")] -> Just x
-                                _         -> Nothing
-
-        winset = fromMaybe initialWinset $ do
-                    s                    <- serializedWinset
-                    ws                   <- maybeRead reads s
-                    return . W.ensureTags layout (workspaces xmc)
-                           $ W.mapLayout (fromMaybe layout . maybeRead lreads) ws
-        extState = fromMaybe M.empty $ do
-                     dyns                        <- serializedExtstate
-                     vals                        <- maybeRead reads dyns
-                     return . M.fromList . map (second Left) $ vals
 
         cf = XConf
             { display       = dpy
@@ -218,14 +215,24 @@ xmonadNoargs initxmc serializedWinset serializedExtstate = do
 
         st = XState
             { windowset       = initialWinset
-            , numberlockMask   = 0
+            , numberlockMask  = 0
             , mapped          = S.empty
             , waitingUnmap    = M.empty
             , dragging        = Nothing
-            , extensibleState = extState
+            , extensibleState = M.empty
             }
+
     allocaXEvent $ \e ->
         runX cf st $ do
+            -- check for serialized state in a file.
+            serializedSt <- do
+                path <- stateFileName
+                exists <- io (doesFileExist path)
+                if exists then readStateFile initxmc else return Nothing
+
+            -- restore extensibleState if we read it from a file.
+            let extst = maybe M.empty extensibleState serializedSt
+            modify (\s -> s {extensibleState = extst})
 
             setNumlockMask
             grabKeys
@@ -240,6 +247,7 @@ xmonadNoargs initxmc serializedWinset serializedExtstate = do
             -- those windows.  Remove all windows that are no longer top-level
             -- children of the root, they may have disappeared since
             -- restarting.
+            let winset = maybe initialWinset windowset serializedSt
             windows . const . foldr W.delete winset $ W.allWindows winset \\ ws
 
             -- manage the as-yet-unmanaged windows
@@ -290,10 +298,10 @@ handle (KeyEvent {ev_event_type = t, ev_state = m, ev_keycode = code})
 
 -- manage a new window
 handle (MapRequestEvent    {ev_window = w}) = withDisplay $ \dpy -> do
-    wa <- io $ getWindowAttributes dpy w -- ignore override windows
-    -- need to ignore mapping requests by managed windows not on the current workspace
-    managed <- isClient w
-    when (not (wa_override_redirect wa) && not managed) $ do manage w
+    withWindowAttributes dpy w $ \wa -> do -- ignore override windows
+      -- need to ignore mapping requests by managed windows not on the current workspace
+      managed <- isClient w
+      when (not (wa_override_redirect wa) && not managed) $ manage w
 
 -- window destroyed, unmanage it
 -- window gone,      unmanage it
@@ -356,7 +364,11 @@ handle e@(ButtonEvent {ev_window = w,ev_event_type = t,ev_button = b })
 -- True in the user's config.
 handle e@(CrossingEvent {ev_window = w, ev_event_type = t})
     | t == enterNotify && ev_mode   e == notifyNormal
-    = whenX (asks $ focusFollowsMouse . config) (focus w)
+    = whenX (asks $ focusFollowsMouse . config) $ do
+        dpy <- asks display
+        root <- asks theRoot
+        (_, _, w', _, _, _, _, _) <- io $ queryPointer dpy root
+        when (w == w') (focus w)
 
 -- left a window, check if we need to focus root
 handle e@(CrossingEvent {ev_event_type = t})
@@ -367,8 +379,6 @@ handle e@(CrossingEvent {ev_event_type = t})
 -- configure a window
 handle e@(ConfigureRequestEvent {ev_window = w}) = withDisplay $ \dpy -> do
     ws <- gets windowset
-    wa <- io $ getWindowAttributes dpy w
-
     bw <- asks (borderWidth . config)
 
     if M.member w (floating ws)
@@ -382,7 +392,7 @@ handle e@(ConfigureRequestEvent {ev_window = w}) = withDisplay $ \dpy -> do
                     , wc_sibling      = ev_above e
                     , wc_stack_mode   = ev_detail e }
                 when (member w ws) (float w)
-        else io $ allocaXEvent $ \ev -> do
+        else withWindowAttributes dpy w $ \wa -> io $ allocaXEvent $ \ev -> do
                  setEventType ev configureNotify
                  setConfigureEvent ev w w
                      (wa_x wa) (wa_y wa) (wa_width wa)
@@ -416,7 +426,7 @@ handle e = broadcastMessage e -- trace (eventName e) -- ignoring
 scan :: Display -> Window -> IO [Window]
 scan dpy rootw = do
     (_, _, ws) <- queryTree dpy rootw
-    filterM ok ws
+    filterM (\w -> ok w `E.catch` skip) ws
   -- TODO: scan for windows that are either 'IsViewable' or where WM_STATE ==
   -- Iconic
   where ok w = do wa <- getWindowAttributes dpy w
@@ -428,6 +438,9 @@ scan dpy rootw = do
                   return $ not (wa_override_redirect wa)
                          && (wa_map_state wa == waIsViewable || ic)
 
+        skip :: E.SomeException -> IO Bool
+        skip _ = return False
+
 setNumlockMask :: X ()
 setNumlockMask = do
     dpy <- asks display
