@@ -1,6 +1,6 @@
 {-# LANGUAGE FlexibleContexts #-}
--- Unlike the rest of xmonad, this file is copyright under the terms of the
--- GPL.
+-- Unlike the rest of xmonad, this file is released under the GNU General
+-- Public License version 2 or later.
 
 --
 -- Generates man/xmonad.1 from man/xmonad.1.in by filling the list of
@@ -79,7 +79,7 @@ main = do
     Right template <- getDefaultTemplate Nothing "man"
     writeFile "./man/xmonad.1"
         . (manHeader ++)
-        . writeMan def{ writerStandalone = True, writerTemplate = template }
+        . writeMan def{ writerTemplate = Just template }
         $ parsed
     putStrLn "Documentation created: man/xmonad.1"
 
@@ -92,8 +92,7 @@ main = do
                              "<p>Section: xmonad manual (1)<br/>"++
                              "Updated: "++releaseDate++"</p>"++
                              "<hr/>")]
-            , writerStandalone = True
-            , writerTemplate = template
+            , writerTemplate = Just template
             , writerTableOfContents = True }
         $ parsed
     putStrLn "Documentation created: man/xmonad.1.html"
